


/*
 * GtkScare is a runner for Adrift text adventure games.
 * This program is released under GNU GPL. See the file COPYING.TXT for
 * details.
 * Copyright (c) 2004 Pallav Nawani
 *
 */


/*! 
*
* \file config.h
* 	No description yet
* 	
* 	
* \author Pallav Nawani
* 
*/

#ifndef _IC_CONFIG_H_
#define _IC_CONFIG_H_

/* read configuration data from a file */
void read_config();

/* Write configuration data to a file */
void write_config();


#endif




/*  End of file config.h  */

