


/*
 * GtkScare is a runner for Adrift text adventure games.
 * This program is released under GNU GPL. See the file COPYING.TXT for
 * details.
 * Copyright (c) 2004 Pallav Nawani
 *
 */


#include "gtk/gtk.h"
#include "gtkscare.h"


void
on_main_win_destroy_event              (GtkObject       *object,
                                        gpointer         user_data);

void on_textentry(GtkObject *entry, gpointer user_data);

void
on_open_game_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_restart_game_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_restore_game_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_save_game_as_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_quit_game_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_exit_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_set_bg_color_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_set_display_font_color_activate     (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_set_background_color_activate       (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_set_font_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_clear_screen_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_verbose_mode_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void on_undo_activate (GtkMenuItem *menuitem,
      gpointer user_data);

void on_gameinfo_activate (GtkMenuItem *menuitem,
      gpointer user_data);


void on_log_input_activate (GtkMenuItem *menuitem,
      gpointer user_data);

void on_enable_debugger_activate (GtkMenuItem *menuitem,
      gpointer user_data);

void on_debugger_console_activate (GtkMenuItem *menuitem,
      gpointer user_data);
void
on_usual_commands_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_about_gtkscare_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data);

void
on_fsel_ok_butn_clicked                (GtkButton       *button,
                                        gpointer         user_data);

void
on_fsel_cancel_butn_clicked            (GtkButton       *button,
                                        gpointer         user_data);

void
on_colorsel_ok_clicked                 (GtkButton       *button,
                                        gpointer         user_data);

void
on_colorsel_cancel_clicked             (GtkButton       *button,
                                        gpointer         user_data);

void
on_fontsel_ok_clicked                  (GtkButton       *button,
                                        gpointer         user_data);

void
on_fontsel_cancel_clicked              (GtkButton       *button,
                                        gpointer         user_data);

void
on_fontsel_apply_clicked               (GtkButton       *button,
                                        gpointer         user_data);

void
on_button3_clicked                     (GtkButton       *button,
                                        gpointer         user_data);

void
on_dlg_yes_butn_clicked                (GtkButton       *button,
                                        gpointer         user_data);

void
on_dlg_no_butn_clicked                 (GtkButton       *button,
                                        gpointer         user_data);

gboolean on_widget_delete( GtkButton *button,
      GdkEvent *event,
      gpointer user_data);


gboolean on_key_press(GtkEntry *entry,
      GdkEventKey *kevent,
      gpointer user_data);


void
on_hintbutton_clicked                  (GtkButton       *button,
                                        gpointer         user_data);

void
on_solbutton_clicked                   (GtkButton       *button,
                                        gpointer         user_data);

void
on_closebutton_clicked                 (GtkButton       *button,
      gpointer user_data);
  
