


/*
 * GtkScare is a runner for Adrift text adventure games.
 * This program is released under GNU GPL. See the file COPYING.TXT for
 * details.
 * Copyright (c) 2004 Pallav Nawani
 *
 */



/*
 * DO NOT EDIT THIS FILE - it is generated by Glade.
 */

#include "gtkscare.h"

// Main window
GtkWidget* create_main_win (Gtk_sdata *idata);

// Creates a confirmation dialog
GtkWidget* create_confirmation_dlg(Gtk_sdata *idata, char *prompt);

GtkWidget* create_fileselection(Gtk_sdata *idata, int mode);
GtkWidget* create_colorseldlg (Gtk_sdata *idata);
GtkWidget* create_fontseldlg (Gtk_sdata *idata);

// A generic message box
GtkWidget* create_about_dlg (char *boxname, char *blurb);

// Creates a window to display hints
GtkWidget* create_hint_window (Gtk_sdata *idata);

// About box
GtkWidget* create_aboutdlg (Gtk_sdata *idata);

// Displays game readme text
GtkWidget* create_game_text_window (char *string);

// Creates the debugger console
GtkWidget* create_debug_window (void);


