


/*
 * GtkScare is a runner for Adrift text adventure games.
 * This program is released under GNU GPL. See the file COPYING.TXT for
 * details.
 * Copyright (c) 2004 Pallav Nawani
 *
 */


#include <gtk/gtk.h>
#include "gtkscare.h"


gboolean
on_debug_kill_event (GtkWidget *widget,
      GdkEventKey *event,
      gpointer user_data);

void
on_dbg_gamebutton_clicked              (GtkButton       *button,
                                        gpointer         user_data);

void
on_dbg_playerbutton_clicked            (GtkButton       *button,
                                        gpointer         user_data);

void
on_dbg_roombutton_clicked              (GtkButton       *button,
                                        gpointer         user_data);

void
on_dbg_objectbutton_clicked            (GtkButton       *button,
                                        gpointer         user_data);

void
on_dbg_npcbutton_clicked               (GtkButton       *button,
                                        gpointer         user_data);

void
on_dbg_taskbutton_clicked              (GtkButton       *button,
                                        gpointer         user_data);

void
on_dbg_eventbutton_clicked             (GtkButton       *button,
                                        gpointer         user_data);

void
on_dbg_variablebutton_clicked          (GtkButton       *button,
                                        gpointer         user_data);

void
on_dbg_helpbutton_clicked              (GtkButton       *button,
                                        gpointer         user_data);

void
on_dbg_closebutton_clicked             (GtkButton       *button,
                                        gpointer         user_data);

void
on_debugentry_activate                 (GtkEntry        *entry,
                                        gpointer         user_data);

gboolean
on_debugentry_key_press_event          (GtkWidget       *widget,
                                        GdkEventKey     *event,
                                        gpointer         user_data);
